LiveLoad.Reporter.Markdown is a very very simple Markdown-based reporter
that can be used to just inspect the outputs of a run in a basic markdown report.
Summary
Functions
Render a markdown report from the result of LiveLoad.run/1.
Functions
@spec render!( results :: %{required(LiveLoad.Scenario.t()) => LiveLoad.scenario_result()} ) :: binary()
Render a markdown report from the result of LiveLoad.run/1.
Returns a binary/0 which can then be used to write to a file, served over a web server, etc.