A single time bucket in the overall scenario's time series metrics.
:offset_msis the millisecond offset from the start of the scenario's run.:active_usersis the number of concurrently running scenario users that were running within this bucket .:node_countindicates how many nodes contributed data to this bucket. On the individual node breakdowns this will benil.
Summary
Types
@type t() :: %LiveLoad.Result.Bucket{ active_users: non_neg_integer(), counters: %{required(String.t()) => LiveLoad.Result.DimensionedCounter.t()}, histograms: %{ required(String.t()) => LiveLoad.Result.DimensionedHistogram.t() }, node_count: pos_integer() | nil, offset_ms: non_neg_integer() }