A result for a scenario run, including summarized user results, aggregated counters, histograms, and bucketed time series data.
Summary
Types
A sample of a failed scenario run to allow debugging and understanding of why users may have failed their scenario.
Types
@type failure_sample() :: %{ kind: :throw | :error | :exit, reason_inspect: String.t(), stacktrace: [String.t()], monotonic_time: integer(), user_id: term() }
A sample of a failed scenario run to allow debugging and understanding of why users may have failed their scenario.
@type t() :: %LiveLoad.Result.ScenarioResult{ counters: %{required(String.t()) => LiveLoad.Result.DimensionedCounter.t()}, duration_ms: pos_integer(), failure_samples: %{required(String.t()) => [failure_sample()]}, histograms: %{ required(String.t()) => LiveLoad.Result.DimensionedHistogram.t() }, time_series: [LiveLoad.Result.Bucket.t()], users: LiveLoad.Result.Users.t() }