LiveLoad.Scenario.Throttle.Parallelism provides a throttle that ensures only a certain amount of parallel executions at any given time.
Example
LiveLoad.Scenario.Throttle.Parallelism.new(:my_throttle, 5)
Summary
Functions
Creates a new LiveLoad.Scenario.Throttle.Parallelism with the given name and concurrency limit.
Types
Functions
@spec new(name :: atom(), limit :: pos_integer()) :: t()
Creates a new LiveLoad.Scenario.Throttle.Parallelism with the given name and concurrency limit.