LiveLoad.Scenario.Throttle.Ramp defines the options for creating a gradual ramp up on a throttle.
Summary
Types
The duration for a ramp up to last.
The interval between steps of steps to take while ramping up.
Options for creating a ramp.
The number of steps to take while ramping up.
The target to ramp up on the throttle.
Types
@type duration_opt() :: {:duration, pos_integer()}
The duration for a ramp up to last.
This option is mutually exclusive with steps_opt/0 and interval_opt/0, which define a step based ramp up.
@type interval_opt() :: {:interval, pos_integer()}
The interval between steps of steps to take while ramping up.
This option requires the steps_opt/0 option to be set as well.
This option is mutually exclusive with duration_opt/0, which defines a duration based ramp up.
@type option() :: duration_opt() | steps_opt() | interval_opt()
Options for creating a ramp.
@type steps_opt() :: {:steps, pos_integer()}
The number of steps to take while ramping up.
This option requires the interval_opt/0 option to be set as well.
This option is mutually exclusive with duration_opt/0, which defines a duration based ramp up.
@type target() :: pos_integer()
The target to ramp up on the throttle.